import os
import pytest

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
IMGCALC_DATA_DIR = os.path.join(DATA_DIR, "imagecalc")

RTREE_NOT_AVAIL = False
try:
    import rtree
except ImportError:
    RTREE_NOT_AVAIL = True

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True


def test_count_pxls_of_val_band1():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    val_counts = rsgislib.imagecalc.count_pxls_of_val(
        input_img, [1, 2, 3, 4], img_band=1
    )
    assert (
        (val_counts[0] == 614)
        and (val_counts[1] == 612)
        and (val_counts[2] == 618)
        and (val_counts[3] == 656)
    )


def test_count_pxls_of_val_band1_sel_vals():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    val_counts = rsgislib.imagecalc.count_pxls_of_val(input_img, [2, 1], img_band=1)
    assert (val_counts[0] == 612) and (val_counts[1] == 614)


def test_count_pxls_of_val_all_bands():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    val_counts = rsgislib.imagecalc.count_pxls_of_val(
        input_img, [1, 2, 3, 4], img_band=None
    )
    assert (
        (val_counts[0] == 1890)
        and (val_counts[1] == 1868)
        and (val_counts[2] == 1861)
        and (val_counts[3] == 1881)
    )


def test_get_unique_values():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    unq_vals = rsgislib.imagecalc.get_unique_values(input_img, img_band=1)
    for val in [1, 2, 3, 4]:
        if val not in unq_vals:
            assert False
    assert True


def test_are_imgs_equal_true():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(input_img, input_img)
    assert img_eq


def test_are_imgs_equal_false():
    import rsgislib.imagecalc

    in_ref_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    in_cmp_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls_v2.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(in_ref_img, in_cmp_img)
    assert not img_eq


def test_are_img_bands_equal_true():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        input_img, 1, input_img, 1
    )
    assert img_eq


def test_are_img_bands_equal_dif_bands_false():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        input_img, 1, input_img, 2
    )
    assert not img_eq


def test_are_img_bands_equal_false():
    import rsgislib.imagecalc

    in_ref_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls.kea")
    in_cmp_img = os.path.join(IMGCALC_DATA_DIR, "test_int_pxls_v2.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        in_ref_img, 1, in_cmp_img, 1
    )
    assert not img_eq


def test_band_maths_sgl_band(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    band_def_seq = list()
    band_def_seq.append(
        rsgislib.imagecalc.BandDefn(band_name="Blue", input_img=input_img, img_band=1)
    )
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_b1.kea")
    rsgislib.imagecalc.band_math(
        output_img, "Blue", "KEA", rsgislib.TYPE_16UINT, band_defs=band_def_seq
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        input_img, 1, output_img, 1
    )
    assert img_eq


def test_band_maths_multi_band(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    ref_ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    band_def_seq = list()
    band_def_seq.append(
        rsgislib.imagecalc.BandDefn(band_name="red", input_img=input_img, img_band=3)
    )
    band_def_seq.append(
        rsgislib.imagecalc.BandDefn(band_name="nir", input_img=input_img, img_band=8)
    )
    output_img = os.path.join(tmp_path, "ndvi_test_band_maths.kea")
    exp = "(nir-red)/(nir+red)"
    rsgislib.imagecalc.band_math(
        output_img, exp, "KEA", rsgislib.TYPE_32FLOAT, band_defs=band_def_seq
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_ndvi_img, 1, output_img, 1
    )
    assert img_eq


def test_band_maths_binary_out(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    ref_ndvi_cats_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea"
    )
    band_def_seq = list()
    band_def_seq.append(
        rsgislib.imagecalc.BandDefn(band_name="ndvi", input_img=input_img, img_band=1)
    )
    output_img = os.path.join(tmp_path, "ndvi_cats_test_band_maths.kea")
    exp = "ndvi>0.95?1:ndvi>0.85?2:ndvi>0.75?3:0"
    rsgislib.imagecalc.band_math(
        output_img, exp, "KEA", rsgislib.TYPE_8UINT, band_defs=band_def_seq
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_ndvi_cats_img, 1, output_img, 1
    )
    assert img_eq


def test_band_maths_exp_err(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    band_def_seq = list()
    band_def_seq.append(
        rsgislib.imagecalc.BandDefn(band_name="ndvi", input_img=input_img, img_band=1)
    )
    output_img = os.path.join(tmp_path, "ndvi_cats_test_err.kea")
    exp = "ndvi>0.5?1:ndvi>0.75?2:0?"
    with pytest.raises(Exception):
        rsgislib.imagecalc.band_math(
            output_img, exp, "KEA", rsgislib.TYPE_32FLOAT, band_defs=band_def_seq
        )


def test_image_band_math_multi_band(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    ref_ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    output_img = os.path.join(tmp_path, "ndvi_test_imageband_math.kea")
    exp = "(b8-b3)/(b8+b3)"
    rsgislib.imagecalc.image_band_math(
        input_img, output_img, exp, "KEA", rsgislib.TYPE_32FLOAT
    )
    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_ndvi_img, 1, output_img, 1
    )
    assert img_eq


def test_image_math_binary_out(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    ref_ndvi_cats_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea"
    )
    output_img = os.path.join(tmp_path, "ndvi_cats_test_image_math.kea")
    exp = "b1>0.95?1:b1>0.85?2:b1>0.75?3:0"
    rsgislib.imagecalc.image_math(
        input_img, output_img, exp, "KEA", rsgislib.TYPE_8UINT
    )
    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_ndvi_cats_img, 1, output_img, 1
    )
    assert img_eq


def test_buffer_img_pxl_vals(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    output_img = os.path.join(tmp_path, "test_buf_vals.kea")

    rsgislib.imagecalc.buffer_img_pxl_vals(
        input_img,
        output_img,
        pxl_vals=[2, 3],
        buf_thres=100,
        tmp_dir=tmp_path,
        gdalformat="KEA",
        img_band=1,
        unit_geo=True,
    )

    assert os.path.exists(output_img)


def test_calc_dist_to_img_vals_sgl_val_geo(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_val2_geo.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals.kea")
    rsgislib.imagecalc.calc_dist_to_img_vals(
        input_img,
        output_img,
        2,
        img_band=1,
        gdalformat="KEA",
        max_dist=100,
        no_data_val=0,
        unit_geo=True,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert img_eq


def test_calc_dist_to_img_vals_sgl_val_pxl(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_val2_pxl.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals.kea")
    rsgislib.imagecalc.calc_dist_to_img_vals(
        input_img,
        output_img,
        2,
        img_band=1,
        gdalformat="KEA",
        max_dist=10,
        no_data_val=0,
        unit_geo=False,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert img_eq


def test_calc_dist_to_img_vals_sgl_val_pxl_out_no_data(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals.kea")
    rsgislib.imagecalc.calc_dist_to_img_vals(
        input_img,
        output_img,
        2,
        img_band=1,
        gdalformat="KEA",
        max_dist=10,
        no_data_val=0,
        out_no_data_val=11,
        unit_geo=False,
    )


def test_calc_dist_to_img_vals_multi_val_geo(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_vals23_geo.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals.kea")
    rsgislib.imagecalc.calc_dist_to_img_vals(
        input_img,
        output_img,
        [2, 3],
        img_band=1,
        gdalformat="KEA",
        max_dist=100,
        no_data_val=0,
        unit_geo=True,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert img_eq


def test_calc_dist_to_img_vals_multi_val_pxl(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_vals23_pxl.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals.kea")
    rsgislib.imagecalc.calc_dist_to_img_vals(
        input_img,
        output_img,
        [2, 3],
        img_band=1,
        gdalformat="KEA",
        max_dist=10,
        no_data_val=0,
        unit_geo=False,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert img_eq


def test_calc_dist_to_img_vals_tiled_sgl_val_geo(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_val2_geo.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals_tiled.kea")
    tmp_dist_dir = os.path.join(tmp_path, "dist_tmp")
    rsgislib.imagecalc.calc_dist_to_img_vals_tiled(
        input_img,
        output_img,
        2,
        img_band=1,
        max_dist=100,
        no_data_val=0,
        gdalformat="KEA",
        unit_geo=True,
        tmp_dir=tmp_dist_dir,
        tile_size=250,
        n_cores=1,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert prop_match > 0.99


def test_calc_dist_to_img_vals_tiled_sgl_val_pxl(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_val2_pxl.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals_tiled.kea")
    tmp_dist_dir = os.path.join(tmp_path, "dist_tmp")
    rsgislib.imagecalc.calc_dist_to_img_vals_tiled(
        input_img,
        output_img,
        2,
        img_band=1,
        max_dist=10,
        no_data_val=0,
        gdalformat="KEA",
        unit_geo=False,
        tmp_dir=tmp_dist_dir,
        tile_size=250,
        n_cores=1,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert prop_match > 0.99


def test_calc_dist_to_img_vals_tiled_multi_val_geo(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_vals23_geo.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals_tiled.kea")
    tmp_dist_dir = os.path.join(tmp_path, "dist_tmp")
    rsgislib.imagecalc.calc_dist_to_img_vals_tiled(
        input_img,
        output_img,
        [2, 3],
        img_band=1,
        max_dist=100,
        no_data_val=0,
        gdalformat="KEA",
        unit_geo=True,
        tmp_dir=tmp_dist_dir,
        tile_size=250,
        n_cores=1,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert prop_match > 0.99


def test_calc_dist_to_img_vals_tiled_multi_val_pxl(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_dist2_vals23_pxl.kea"
    )

    output_img = os.path.join(tmp_path, "test_calc_dist_to_img_vals_tiled.kea")
    tmp_dist_dir = os.path.join(tmp_path, "dist_tmp")
    rsgislib.imagecalc.calc_dist_to_img_vals_tiled(
        input_img,
        output_img,
        [2, 3],
        img_band=1,
        max_dist=10,
        no_data_val=0,
        gdalformat="KEA",
        unit_geo=False,
        tmp_dir=tmp_dist_dir,
        tile_size=250,
        n_cores=1,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_img_bands_equal(
        ref_img, 1, output_img, 1
    )
    assert prop_match > 0.99


def test_calc_prop_true_exp_no_vld_msk():
    import rsgislib.imagecalc

    ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")

    band_defns = [rsgislib.imagecalc.BandDefn("ndvi", ndvi_img, 1)]
    prop = rsgislib.imagecalc.calc_prop_true_exp("ndvi>0.5?1:0", band_defns)

    assert abs(prop - 0.855) < 0.001


def test_calc_prop_true_exp_vld_msk():
    import rsgislib.imagecalc

    ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    vld_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")

    band_defns = [rsgislib.imagecalc.BandDefn("ndvi", ndvi_img, 1)]
    prop = rsgislib.imagecalc.calc_prop_true_exp("ndvi>0.5?1:0", band_defns, vld_img)

    assert abs(prop - 0.655) < 0.001


def test_calc_band_percentile():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    percent_val = rsgislib.imagecalc.calc_band_percentile(input_img, 0.5, 0)

    assert ((percent_val[0] - 43) < 1) and ((percent_val[5] - 458) < 1)


def test_calc_img_rescale_sgl_img(tmp_path):
    import rsgislib.imagecalc

    rescale_ndvi_ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_g100_o100_int.kea"
    )
    ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    print(ndvi_img)

    output_img = os.path.join(tmp_path, "rescaled_ndvi_img.kea")

    rsgislib.imagecalc.calc_img_rescale(
        ndvi_img, output_img, "KEA", rsgislib.TYPE_16INT, -999, 0, 1, 999, 100, 100
    )

    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, rescale_ndvi_ref_img
    )
    assert img_eq


def test_calc_img_rescale_multi_imgs(tmp_path):
    import rsgislib.imagecalc

    rescale_ndvi_ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_bs2_g100_o100_int.kea"
    )
    ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")
    print(ndvi_img)

    output_img = os.path.join(tmp_path, "rescaled_ndvi_imgs.kea")

    rsgislib.imagecalc.calc_img_rescale(
        [ndvi_img, ndvi_img],
        output_img,
        "KEA",
        rsgislib.TYPE_16INT,
        -999,
        0,
        1,
        999,
        100,
        100,
    )

    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(
        output_img, rescale_ndvi_ref_img
    )
    assert img_eq


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_perform_image_pca(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    output_img = os.path.join(tmp_path, "pca_result_img.kea")
    out_eigen_vec_file = os.path.join(tmp_path, "pca_eign_vec.mtxt")

    rsgislib.imagecalc.perform_image_pca(
        input_img,
        output_img,
        out_eigen_vec_file,
        n_comps=None,
        pxl_n_sample=100,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_32FLOAT,
        no_data_val=None,
        calc_stats=True,
    )

    assert os.path.exists(output_img) and os.path.exists(out_eigen_vec_file)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_perform_image_mnf(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    output_img = os.path.join(tmp_path, "pca_result_img.kea")
    tmp_dir = os.path.join(tmp_path, "tmp")

    rsgislib.imagecalc.perform_image_mnf(
        input_img,
        output_img,
        n_comps=None,
        pxl_n_sample=100,
        in_img_no_data=0,
        tmp_dir=tmp_dir,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_32FLOAT,
        calc_stats=True,
    )

    assert os.path.exists(output_img)


def test_recode_int_raster(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")

    output_img = os.path.join(tmp_path, "recoded_ndvi_cats.kea")

    recode_dict = dict()
    recode_dict[1] = 100
    recode_dict[2] = 200
    recode_dict[3] = 300

    rsgislib.imagecalc.recode_int_raster(
        input_img,
        output_img,
        recode_dict,
        keep_vals_not_in_dict=True,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_16UINT,
    )
    assert os.path.exists(output_img)


def test_all_bands_equal_to(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.all_bands_equal_to(
        input_img, output_img, 0, 1, 2, "KEA", rsgislib.TYPE_8UINT
    )
    assert os.path.exists(output_img)


def test_kmeans_clustering(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_file = os.path.join(tmp_path, "out_file")
    rsgislib.imagecalc.kmeans_clustering(
        input_img,
        out_file,
        10,
        20,
        100,
        True,
        0.0025,
        rsgislib.INITCLUSTER_DIAGONAL_FULL_ATTACH,
    )

    out_ext_file = os.path.join(tmp_path, "out_file.gmtxt")
    assert os.path.exists(out_ext_file)


def test_isodata_clustering(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    out_file = os.path.join(tmp_path, "out_file")
    rsgislib.imagecalc.isodata_clustering(
        input_img,
        out_file,
        10,
        20,
        100,
        True,
        0.0025,
        rsgislib.INITCLUSTER_DIAGONAL_FULL_ATTACH,
        2,
        5,
        5,
        5,
        8,
        18,
    )
    out_ext_file = os.path.join(tmp_path, "out_file.gmtxt")
    assert os.path.exists(out_ext_file)


def test_image_pixel_column_summary(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    stats_to_calc = rsgislib.imagecalc.StatsSummary()
    stats_to_calc.calc_min = True
    stats_to_calc.calc_max = True
    stats_to_calc.calc_sum = True
    stats_to_calc.calc_mean = True
    stats_to_calc.calc_stdev = True
    stats_to_calc.calc_median = True
    stats_to_calc.calc_mode = True
    rsgislib.imagecalc.image_pixel_column_summary(
        input_img, output_img, stats_to_calc, "KEA", rsgislib.TYPE_32FLOAT, 0.0, True
    )
    assert os.path.exists(output_img)


def test_get_img_band_stats_in_env():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_wgs84.kea")
    stats = rsgislib.imagecalc.get_img_band_stats_in_env(
        input_img, 1, 0, -4.0932, -4.0774, 52.4033, 52.4229
    )

    print(stats)
    assert (
        ((stats[0] - 30) < 1)
        and ((stats[1] - 1066) < 1)
        and ((stats[2] - 74.6039) < 1)
        and ((stats[3] - 34.9773) < 1)
        and ((stats[4] - 1904637.0) < 2)
    )


def test_get_img_band_mode_in_env_0():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_wgs84.kea")
    mode = rsgislib.imagecalc.get_img_band_mode_in_env(
        input_img, 1, 1, -999, -4.0932, -4.0774, 52.4033, 52.4229
    )

    assert mode[0] == 0


def test_get_img_band_mode_in_env_2():
    import rsgislib.imagecalc

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_wgs84.kea")
    mode = rsgislib.imagecalc.get_img_band_mode_in_env(
        input_img, 1, 1, -999, -4.0154, -3.9880, 52.4004, 52.4257
    )

    print(mode)
    assert mode[0] == 2


def test_calc_multi_img_band_stats(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calc_multi_img_band_stats(
        [input_img, input_img],
        output_img,
        rsgislib.SUMTYPE_MEAN,
        "KEA",
        rsgislib.TYPE_32FLOAT,
        0,
        True,
    )

    assert os.path.exists(output_img)


def test_get_img_band_min_max():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    min, max = rsgislib.imagecalc.get_img_band_min_max(input_img, 1, True, 0)
    assert min == 1 and max == 1066


# TODO rsgislib.imagecalc.get_img_sum_stats_in_pxl
# TODO rsgislib.imagecalc.get_img_idx_for_stat
# TODO rsgislib.imagecalc.identify_min_pxl_value_in_win


def test_calc_img_mean_in_mask():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    in_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    val = rsgislib.imagecalc.calc_img_mean_in_mask(
        input_img,
        in_msk_img,
        msk_val=1,
        img_bands=[1, 2, 3],
        no_data_val=0,
        use_no_data=True,
    )
    assert (val - 71.5) < 0.1


def test_calc_img_basic_stats_for_ref_region(tmp_path):
    import rsgislib.imagecalc

    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    in_stats_imgs = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calc_img_basic_stats_for_ref_region(
        in_ref_img, [in_stats_imgs], output_img, gdalformat="KEA"
    )

    assert os.path.exists(output_img)


@pytest.mark.skipif(RTREE_NOT_AVAIL, reason="rtree dependency not available")
def test_calc_fill_regions_knn(tmp_path):
    import rsgislib.imagecalc

    in_ref_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_subset.kea"
    )
    in_fill_regions_img = os.path.join(
        IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats_subset.kea"
    )
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calc_fill_regions_knn(
        in_ref_img,
        0,
        in_fill_regions_img,
        0,
        output_img,
        k=5,
        summary=rsgislib.SUMTYPE_MODE,
        gdalformat="KEA",
        datatype=rsgislib.TYPE_16INT,
    )
    assert os.path.exists(output_img)


def test_image_pixel_linear_fit(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    band_values = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.image_pixel_linear_fit(
        input_img, output_img, "KEA", band_values, 0, True
    )
    assert os.path.exists(output_img)


def test_calculate_img_band_rmse():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imagecalc.calculate_img_band_rmse(input_img, 1, input_img, 2)


def test_correlation_window(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.correlation_window(
        input_img, output_img, 5, 1, 2, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_calc_mask_img_pxl_val_prob(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    in_msk_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calc_mask_img_pxl_val_prob(
        input_img, [1, 2, 3], in_msk_img, 2, output_img, "KEA", [5, 5, 5], False, False
    )
    assert os.path.exists(output_img)


def test_calc_img_difference(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.calc_img_difference(
        input_img, input_img, output_img, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_histogram(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    in_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    output_file = os.path.join(tmp_path, "out_img.txt")
    rsgislib.imagecalc.histogram(
        input_img, in_msk_img, output_file, 1, 1, 10, True, 0, 0
    )

    assert os.path.exists(output_file)


def test_get_histogram():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    rsgislib.imagecalc.get_histogram(input_img, 1, 1, True, 0, 0)


def test_get_2d_img_histogram(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.get_2d_img_histogram(
        input_img,
        input_img,
        output_img,
        "KEA",
        1,
        2,
        100,
        1,
        1,
        1000,
        1000,
        1,
        1,
        0,
        0,
        False,
    )
    assert os.path.exists(output_img)


def test_calc_histograms_for_msk_vals():
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    in_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")

    rsgislib.imagecalc.calc_histograms_for_msk_vals(
        input_img, 1, in_msk_img, 1, 1, 1000, 1, msk_vals=None
    )


def test_normalise_image_band(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.normalise_image_band(input_img, 1, output_img, gdal_format="KEA")
    assert os.path.exists(output_img)


def test_rescale_img_pxl_vals(tmp_path):
    import rsgislib.imagecalc

    band_rescale_objs = list()
    band_rescale_objs.append(
        rsgislib.imagecalc.ImageBandRescale(
            band=1,
            in_min=1.0,
            in_max=1000.0,
            no_data_val=0,
            out_min=200.0,
            out_max=500.0,
            out_no_data=-999.0,
        )
    )
    band_rescale_objs.append(
        rsgislib.imagecalc.ImageBandRescale(
            band=2,
            in_min=1.0,
            in_max=1000.0,
            no_data_val=0,
            out_min=200.0,
            out_max=500.0,
            out_no_data=-999.0,
        )
    )
    band_rescale_objs.append(
        rsgislib.imagecalc.ImageBandRescale(
            band=3,
            in_min=1.0,
            in_max=1000.0,
            no_data_val=0,
            out_min=200.0,
            out_max=500.0,
            out_no_data=-999.0,
        )
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.rescale_img_pxl_vals(
        input_img,
        output_img,
        "KEA",
        rsgislib.TYPE_32FLOAT,
        band_rescale_objs,
        trim_to_limits=True,
    )
    assert os.path.exists(output_img)


def test_mahalanobis_dist_filter(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.mahalanobis_dist_filter(
        input_img, output_img, 5, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)


def test_mahalanobis_dist_to_img_filter(tmp_path):
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imagecalc.mahalanobis_dist_to_img_filter(
        input_img, output_img, 5, "KEA", rsgislib.TYPE_32FLOAT
    )
    assert os.path.exists(output_img)



import os
import rsgislib

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_apply_median_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_median_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_mean_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_mean_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_min_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_min_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_max_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_max_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_mode_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_mode_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_stddev_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_stddev_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_range_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_range_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_mean_diff_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_mean_diff_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_mean_diff_abs_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_mean_diff_abs_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_total_diff_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_total_diff_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_median_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_median_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_total_diff_abs_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_total_diff_abs_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_coeff_of_var_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_coeff_of_var_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_total_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_total_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_norm_var_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_norm_var_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_norm_var_sqrt_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_norm_var_sqrt_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_norm_var_ln_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_norm_var_ln_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_texture_var_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_texture_var_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_kuwahara_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_kuwahara_filter(
        input_img, output_img, 3, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_sobel_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_sobel_filter(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_sobel_x_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_sobel_x_filter(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_sobel_y_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_sobel_y_filter(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_prewitt_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_prewitt_filter(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_prewitt_x_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_prewitt_x_filter(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_prewitt_y_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_prewitt_y_filter(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_gaussian_smooth_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_gaussian_smooth_filter(
        input_img, output_img, 3, 1, 1, 45, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_gaussian_1st_deriv_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_gaussian_1st_deriv_filter(
        input_img, output_img, 3, 1, 1, 45, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_gaussian_2nd_deriv_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_gaussian_2nd_deriv_filter(
        input_img, output_img, 3, 1, 1, 45, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_laplacian_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_laplacian_filter(
        input_img, output_img, 3, 1, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_apply_lee_filter(tmp_path):
    import rsgislib.imagefilter

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.apply_lee_filter(
        input_img, output_img, 3, 2, "KEA", rsgislib.TYPE_16UINT
    )

    assert os.path.exists(output_img)


def test_leung_malik_filter_bank(tmp_path):
    import rsgislib.imagefilter
    import glob

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    out_img_base = os.path.join(tmp_path, "filter_output")
    rsgislib.imagefilter.leung_malik_filter_bank(
        input_img, out_img_base, "KEA", "kea", rsgislib.TYPE_16UINT
    )

    imgs = glob.glob("{}*.kea".format(out_img_base))

    assert len(imgs) == 48


def test_apply_filters(tmp_path):
    import rsgislib.imagefilter
    import glob

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    out_img_base = os.path.join(tmp_path, "filter_output")

    filters = []
    filters.append(
        rsgislib.imagefilter.FilterParameters(
            filter_type="GaussianSmooth",
            file_ending="gausmooth",
            size=3,
            stddev_x=1.0,
            stddev_y=1,
            angle=0.0,
        )
    )
    filters.append(
        rsgislib.imagefilter.FilterParameters(
            filter_type="Gaussian1st",
            file_ending="gau1st",
            size=3,
            stddev_x=1.0,
            stddev_y=1,
            angle=0.0,
        )
    )
    filters.append(
        rsgislib.imagefilter.FilterParameters(
            filter_type="Gaussian2nd",
            file_ending="gau2st",
            size=3,
            stddev_x=1.0,
            stddev_y=1,
            angle=0.0,
        )
    )
    # Apply filters
    rsgislib.imagefilter.apply_filters(
        input_img, out_img_base, filters, "KEA", "kea", rsgislib.TYPE_16UINT
    )

    imgs = glob.glob("{}*.kea".format(out_img_base))

    assert len(imgs) == 3


def test_perform_tiled_img_multi_filter(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_insts = [
        rsgislib.imagefilter.tiledfilter.RSGISMedianFilter(
            3, "KEA", rsgislib.TYPE_16UINT
        ),
        rsgislib.imagefilter.tiledfilter.RSGISMeanFilter(
            3, "KEA", rsgislib.TYPE_16UINT
        ),
    ]

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_imgs = [
        os.path.join(tmp_path, "filter_output1.kea"),
        os.path.join(tmp_path, "filter_output2.kea"),
    ]
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_multi_filter(
        input_img,
        output_imgs,
        filter_insts,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_imgs[0]) and os.path.exists(output_imgs[1])


def test_perform_tiled_img_filter_median(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISMedianFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_mean(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISMeanFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_min(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISMinFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_max(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISMaxFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_mode(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISModeFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_stddev(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISStdDevFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_range(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISRangeFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_mean_dif(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISMeanDiffFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_mean_diff_abs(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISMeanDiffAbsFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_total_diff(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISTotalDiffFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_total_diff_abs(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISTotalDiffAbsFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_coeff_of_var(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISCoeffOfVarFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_total(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISTotalFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_norm_var(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISNormVarFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_norm_var_sqrt(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISNormVarSqrtFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_norm_var_ln(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISNormVarLnFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_texture_var(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISTextureVarFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_kuwahara(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISKuwaharaFilter(
        3, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_sobel(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISSobelFilter(
        "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_sobel_x(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISSobelXFilter(
        "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_sobel_y(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISSobelYFilter(
        "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_prewitt(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISPrewittFilter(
        "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_prewitt_x(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISPrewittXFilter(
        "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_prewitt_y(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISPrewittYFilter(
        "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_gaussian_smooth(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISGaussianSmoothFilter(
        3, 1, 1, 0, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_gaussian_1st_deriv(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISGaussian1stDerivFilter(
        3, 1, 1, 0, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_gaussian_2nd_deriv(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISGaussian2ndDerivFilter(
        3, 1, 1, 0, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_laplacian(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISLaplacianFilter(
        3, 1, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)


def test_perform_tiled_img_filter_lee(tmp_path):
    import rsgislib.imagefilter.tiledfilter

    filter_inst = rsgislib.imagefilter.tiledfilter.RSGISLeeFilter(
        3, 2, "KEA", rsgislib.TYPE_16UINT
    )

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "filter_output.kea")
    rsgislib.imagefilter.tiledfilter.perform_tiled_img_filter(
        input_img,
        output_img,
        filter_inst,
        datatype=None,
        gdalformat="KEA",
        tmp_dir="tmp",
        width=250,
        height=250,
        n_cores=1,
    )

    assert os.path.exists(output_img)



#!/usr/bin/env python
"""
The tools.test_images module contains some functions to create images which can be
used for testing other functions or generate example datasets

"""
import rsgislib


def create_random_int_img(
    output_img,
    n_bands,
    x_size,
    y_size,
    out_vals,
    gdalformat="KEA",
    datatype=rsgislib.TYPE_8UINT,
    calc_stats=True,
):
    """
    A function which creates an image with pixels values randomly assigned from the
    list in the out_vals input list.

    :param output_img: The output image file name and path
    :param n_bands: the number of bands in the output image
    :param x_size: the number of pixels in the X axis
    :param y_size: the number of pixels in the Y axis
    :param out_vals: a list of values to be randomly populated within the output image
    :param gdalformat: the output image file format (Default: KEA)
    :param datatype: the output image data type (Default: rsgislib.TYPE_8UINT)
    :param calc_stats: calculate image statistics and pyramids

    """
    import rsgislib.imageutils
    from rios import applier
    import numpy.random

    rsgislib.imageutils.createBlankImagePy(
        output_img,
        n_bands,
        x_size,
        y_size,
        100000,
        500000,
        1,
        1,
        "",
        gdalformat,
        datatype,
        options=[],
        no_data_val=0,
    )

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.image = output_img
    outfiles = applier.FilenameAssociations()
    outfiles.out_image = output_img
    otherargs = applier.OtherInputs()
    otherargs.out_vals = out_vals
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    def _applyPopVals(info, inputs, outputs, otherargs):
        """
        This is an internal rios function
        """
        in_img_shp = inputs.image.shape
        outputs.out_image = numpy.random.choice(otherargs.out_vals, in_img_shp).astype(
            rsgislib.get_numpy_datatype(datatype)
        )

    applier.apply(_applyPopVals, infiles, outfiles, otherargs, controls=aControls)
    if calc_stats:
        import rsgislib.rastergis

        rsgislib.rastergis.pop_rat_img_stats(output_img, True, True, True)



#! /usr/bin/env python
############################################################################
#  shepherdseg.py
#
#  Copyright 2013 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 16/11/2013
# Version: 1.1
#
# History:
# Version 1.0 - Created.
# Version 1.1 - Update to be included into RSGISLib python modules tree.
#
###########################################################################
# Import shutil
import shutil

# Import the rsgislib module
import rsgislib

# Import the image utilities module from rsgislib
import rsgislib.imageutils

# Import the image calc module from rsgislib
import rsgislib.imagecalc

# Import the image segmentation module from rsgislib
import rsgislib.segmentation

# Import the image rastergis module from rsgislib
import rsgislib.rastergis

# Import the python OS module
import os

# Import the collections module
import collections

# import the gdal module
import osgeo.gdal as gdal

# Import the python JSON module
import json


def run_shepherd_segmentation(
    input_img,
    out_clumps_img,
    out_mean_img=None,
    tmp_dir=".",
    gdalformat="KEA",
    calc_stats=True,
    no_stretch=False,
    no_delete=False,
    num_clusters=60,
    min_n_pxls=100,
    dist_thres=100,
    bands=None,
    sampling=100,
    km_max_iter=200,
    process_in_mem=False,
    save_process_stats=False,
    img_stretch_stats="",
    kmeans_centres="",
    img_stats_json_file="",
):
    """
    Utility function to call the segmentation algorithm of Shepherd et al. (2019).

    Shepherd, J. D., Bunting, P., & Dymond, J. R. (2019). Operational Large-Scale
    Segmentation of Imagery Based on Iterative Elimination. Remote Sensing, 11(6), 658.
    http://doi.org/10.3390/rs11060658

    :param input_img: is a string containing the name of the input file.
    :param out_clumps_img: is a string containing the name of the output clump file.
    :param out_mean_img: is the output mean image file (clumps attributed with pixel
                         mean from input image) - pass 'None' to skip creating.
    :param tmp_dir: is a file path for intermediate files (default is current
                    directory).
    :param gdalformat: is a string containing the GDAL format for the output file
                       (default = KEA).
    :param calc_stats: is a bool which specifies that image statistics and pyramids
                       should be built for the output images (default = True)
    :param no_stretch: is a bool which specifies that the input image bands should
                       not be stretched (default = False).
    :param no_delete: is a bool which specifies that the temporary images created
                      during processing should not be deleted once processing has
                      been completed (default = False).
    :param num_clusters: is an int which specifies the number of clusters within the
                         KMeans clustering (default = 60).
    :param min_n_pxls: is an int which specifies the minimum number pixels within a
                       segments (default = 100).
    :param dist_thres: specifies the distance threshold for joining the segments
                       (default = 100, set to large number to turn off this option).
    :param bands: is an array providing a subset of image bands to use (default
                  is None to use all bands).
    :param sampling: specify the subsampling of the image for the data used within
                     the KMeans (default = 100; 1 == no subsampling).
    :param km_max_iter: maximum iterations for KMeans.
    :param process_in_mem: where functions allow it perform processing in memory
                           rather than on disk.
    :param save_process_stats: is a bool which specifies that the image stretch
                               stats and the kMeans centre stats should be saved
                               along with a header.
    :param img_stretch_stats: is a string providing the file name and path for the
                              image stretch stats (Output).
    :param kmeans_centres: is a string providing the file name and path for the
                           KMeans clusters centres (don't include file extension;
                           .gmtxt will be added to the end) (Output).
    :param img_stats_json_file: is a string providing the name and path of a JSON
                                file storing the image spatial extent and
                                img_stretch_stats and kmeans_centres file paths
                                for use by other commands (Output).

    .. code:: python

        from rsgislib.segmentation import shepherdseg

        input_img = 'jers1palsar_stack.kea'
        out_clumps_img = 'jers1palsar_stack_clumps_elim_final.kea'
        out_mean_img = 'jers1palsar_stack_clumps_elim_final_mean.kea'

        shepherdseg.run_shepherd_segmentation(input_img, out_clumps_img,
                                              out_mean_img, min_n_pxls=100)

    """
    import rsgislib.tools.filetools

    if save_process_stats:
        if (
            (img_stretch_stats == "")
            or (kmeans_centres == "")
            or (img_stats_json_file == "")
        ):
            raise rsgislib.RSGISPyException(
                "if image stretch and kmeans centres are to be saved then all "
                "file names (img_stretch_stats, kmeans_centres, img_stats_json_file) "
                "need to be provided."
            )

    basefile = os.path.basename(input_img)
    basename = os.path.splitext(basefile)[0]

    out_file_ext = rsgislib.imageutils.get_file_img_extension(gdalformat)

    createdDIR = False
    if not os.path.isdir(tmp_dir):
        os.makedirs(tmp_dir)
        createdDIR = True

    # Get data type of input image
    input_datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)

    # Select Image Bands if required
    inputImgBands = input_img
    selectBands = False
    if not bands is None:
        print("Subsetting the image bands")
        selectBands = True
        inputImgBands = os.path.join(
            tmp_dir, "{}_bselect.{}".format(basename, out_file_ext)
        )
        rsgislib.imageutils.select_img_bands(
            input_img, inputImgBands, gdalformat, input_datatype, bands
        )

    # Stretch input data if required.
    segmentFile = inputImgBands
    if not no_stretch:
        segmentFile = os.path.join(
            tmp_dir, "{}_stchd.{}".format(basename, out_file_ext)
        )
        strchFile = os.path.join(
            tmp_dir, "{}_stchdonly.{}".format(basename, out_file_ext)
        )
        strchFileOffset = os.path.join(
            tmp_dir, "{}_stchdonly_off.{}".format(basename, out_file_ext)
        )
        strchMaskFile = os.path.join(
            tmp_dir, "{}_stchdmaskonly.{}".format(basename, out_file_ext)
        )

        print("Stretch Input Image")
        rsgislib.imageutils.stretch_img(
            inputImgBands,
            strchFile,
            save_process_stats,
            img_stretch_stats,
            True,
            False,
            gdalformat,
            rsgislib.TYPE_8UINT,
            rsgislib.imageutils.STRETCH_LINEARSTDDEV,
            2,
        )

        print(
            "Add 1 to stretched file to ensure there are no all "
            "zeros (i.e., no data) regions created."
        )
        rsgislib.imagecalc.image_math(
            strchFile, strchFileOffset, "b1+1", gdalformat, rsgislib.TYPE_8UINT
        )

        print("Create Input Image Mask.")
        bandMathBands = [
            rsgislib.imagecalc.BandDefn(
                band_name="b1", input_img=inputImgBands, img_band=1
            )
        ]
        rsgislib.imagecalc.band_math(
            strchMaskFile, "b1==0?1:0", gdalformat, rsgislib.TYPE_8UINT, bandMathBands
        )

        print("Mask stretched Image.")
        rsgislib.imageutils.mask_img(
            strchFileOffset,
            strchMaskFile,
            segmentFile,
            gdalformat,
            rsgislib.TYPE_8UINT,
            0,
            1,
        )

        if not no_delete:
            # Deleting extra files
            rsgislib.tools.filetools.delete_file_with_basename(strchFile)
            rsgislib.tools.filetools.delete_file_with_basename(strchFileOffset)
            rsgislib.tools.filetools.delete_file_with_basename(strchMaskFile)

    # Perform KMEANS
    print("Performing KMeans.")
    outMatrixFile = os.path.join(tmp_dir, "{}_kmeansclusters".format(basename))
    if save_process_stats:
        outMatrixFile = kmeans_centres
    rsgislib.imagecalc.kmeans_clustering(
        segmentFile,
        outMatrixFile,
        num_clusters,
        km_max_iter,
        sampling,
        True,
        0.0025,
        rsgislib.imagecalc.INITCLUSTER_DIAGONAL_FULL_ATTACH,
    )

    # Apply KMEANS
    print("Apply KMeans to image.")
    kMeansFileZones = os.path.join(
        tmp_dir, "{}_kmeans.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.label_pixels_from_cluster_centres(
        segmentFile, kMeansFileZones, outMatrixFile + str(".gmtxt"), True, gdalformat
    )

    # Eliminate Single Pixels
    print("Eliminate Single Pixels.")
    kMeansFileZonesNoSgls = os.path.join(
        tmp_dir, "{}_kmeans_nosgl.{}".format(basename, out_file_ext)
    )
    kMeansFileZonesNoSglsTmp = os.path.join(
        tmp_dir, "{}_kmeans_nosgl_tmp.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.eliminate_single_pixels(
        segmentFile,
        kMeansFileZones,
        kMeansFileZonesNoSgls,
        kMeansFileZonesNoSglsTmp,
        gdalformat,
        process_in_mem,
        True,
    )

    # Clump
    print("Perform clump.")
    initClumpsFile = os.path.join(
        tmp_dir, "{}_clumps.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.clump(
        kMeansFileZonesNoSgls, initClumpsFile, gdalformat, process_in_mem, 0
    )

    # Elimininate small clumps
    print("Eliminate small pixels.")
    elimClumpsFile = os.path.join(
        tmp_dir, "{}_clumps_elim.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.rm_small_clumps_stepwise(
        segmentFile,
        initClumpsFile,
        elimClumpsFile,
        gdalformat,
        False,
        "",
        False,
        process_in_mem,
        min_n_pxls,
        dist_thres,
    )

    # Relabel clumps
    print("Relabel clumps.")
    rsgislib.segmentation.relabel_clumps(
        elimClumpsFile, out_clumps_img, gdalformat, process_in_mem
    )

    # Populate with stats if required.
    if calc_stats:
        print("Calculate image statistics and build pyramids.")
        rsgislib.rastergis.pop_rat_img_stats(out_clumps_img, True, True)

    # Create mean image if required.
    if not (out_mean_img is None):
        rsgislib.segmentation.mean_image(
            input_img, out_clumps_img, out_mean_img, gdalformat, input_datatype
        )
        if calc_stats:
            rsgislib.imageutils.pop_img_stats(out_mean_img, True, 0, True)

    if save_process_stats:
        gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
        geotransform = gdalDS.GetGeoTransform()
        if not geotransform is None:
            xTL = geotransform[0]
            yTL = geotransform[3]

            xRes = geotransform[1]
            yRes = geotransform[5]

            width = gdalDS.RasterXSize * xRes
            if yRes < 0:
                yRes = yRes * (-1)
            height = gdalDS.RasterYSize * yRes
            xBR = xTL + width
            yBR = yTL - height

            xCen = xTL + (width / 2)
            yCen = yBR + (height / 2)

            sceneData = dict()
            sceneData["KCENTRES"] = kmeans_centres + str(".gmtxt")
            sceneData["STRETCHSTATS"] = img_stretch_stats
            sceneData["CENTRE_PT"] = {"X": xCen, "Y": yCen}
            sceneData["BBOX"] = {"XMIN": xTL, "YMIN": yBR, "XMAX": xBR, "YMAX": yTL}

            with open(img_stats_json_file, "w") as outfile:
                json.dump(
                    sceneData,
                    outfile,
                    sort_keys=True,
                    indent=4,
                    separators=(",", ": "),
                    ensure_ascii=False,
                )

        gdalDS = None

    if not no_delete:
        # Deleting extra files
        if not save_process_stats:
            rsgislib.tools.filetools.delete_file_with_basename(
                outMatrixFile + str(".gmtxt")
            )
        rsgislib.tools.filetools.delete_file_with_basename(kMeansFileZones)
        rsgislib.tools.filetools.delete_file_with_basename(kMeansFileZonesNoSgls)
        rsgislib.tools.filetools.delete_file_with_basename(kMeansFileZonesNoSglsTmp)
        rsgislib.tools.filetools.delete_file_with_basename(initClumpsFile)
        rsgislib.tools.filetools.delete_file_with_basename(elimClumpsFile)
        if selectBands:
            rsgislib.tools.filetools.delete_file_with_basename(inputImgBands)
        if not no_stretch:
            rsgislib.tools.filetools.delete_file_with_basename(segmentFile)
        if createdDIR:
            shutil.rmtree(tmp_dir)


def run_shepherd_segmentation_pre_calcd_stats(
    input_img,
    out_clumps_img,
    kmeans_centres,
    img_stretch_stats,
    out_mean_img=None,
    tmp_dir=".",
    gdalformat="KEA",
    calc_stats=True,
    no_stretch=False,
    no_delete=False,
    min_n_pxls=100,
    dist_thres=100,
    bands=None,
    process_in_mem=False,
):
    """
    Utility function to call the segmentation algorithm of Shepherd et al. (2019)
    using pre-calculated stretch stats and KMeans cluster centres.

    Shepherd, J. D., Bunting, P., & Dymond, J. R. (2019). Operational Large-Scale
    Segmentation of Imagery Based on Iterative Elimination. Remote Sensing, 11(6),
    658. http://doi.org/10.3390/rs11060658


    :param input_img: is a string containing the name of the input file.
    :param out_clumps_img: is a string containing the name of the output clump file.
    :param kmeans_centres: is a string providing the file name and path for the
                           KMeans clusters centres (Input)
    :param img_stretch_stats: is a string providing the file name and path for the
                              image stretch stats (Input - not required if
                              no_stretch=True)
    :param out_mean_img: is the output mean image file (clumps attributed with pixel
                         mean from input image) - pass 'None' to skip creating.
    :param tmp_dir: is a file path for intermediate files (default is current
                    directory).
    :param gdalformat: is a string containing the GDAL format for the output file
                       (default = KEA).
    :param calc_stats: is a bool which specifies that image statistics and pyramids
                       should be built for the output images (default = True)
    :param no_stretch: is a bool which specifies that the input image bands should
                       not be stretched (default = False).
    :param no_delete: is a bool which specifies that the temporary images created
                      during processing should not be deleted once processing has
                      been completed (default = False).
    :param min_n_pxls: is an int which specifies the minimum number pixels within
                       a segments (default = 100).
    :param dist_thres: specifies the distance threshold for joining the segments
                       (default = 100, set to large number to turn off this option).
    :param bands: is an array providing a subset of image bands to use (default is
                  None to use all bands).
    :param sampling: specify the subsampling of the image for the data used within
                     the KMeans (default = 100; 1 == no subsampling).
    :param process_in_mem: where functions allow it perform processing in memory
                           rather than on disk.

    .. code:: python

        from rsgislib.segmentation import shepherdseg

        input_img = 'jers1palsar_stack.kea'
        out_clumps_img = 'jers1palsar_stack_clumps_elim_final.kea'
        out_mean_img = 'jers1palsar_stack_clumps_elim_final_mean.kea'
        kmeans_centres = 'jers1palsar_stack_kcentres.gmtxt'
        img_stretch_stats = 'jers1palsar_stack_stchstats.txt'

        shepherdseg.run_shepherd_segmentation_pre_calcd_stats(input_img, out_clumps_img,
                                                              kmeans_centres,
                                                              img_stretch_stats,
                                                              out_mean_img,
                                                              min_n_pxls=100)

    """
    import rsgislib.tools.filetools

    if not no_stretch:
        if (img_stretch_stats == "") or (img_stretch_stats is None):
            raise rsgislib.RSGISPyException("A stretch stats file must be provided")

    basefile = os.path.basename(input_img)
    basename = os.path.splitext(basefile)[0]

    out_file_ext = rsgislib.imageutils.get_file_img_extension(gdalformat)

    createdDIR = False
    if not os.path.isdir(tmp_dir):
        os.makedirs(tmp_dir)
        createdDIR = True

    # Get data type of input image
    gdalDS = gdal.Open(input_img, gdal.GA_ReadOnly)
    input_datatype = rsgislib.get_rsgislib_datatype(
        gdal.GetDataTypeName(gdalDS.GetRasterBand(1).DataType)
    )
    gdalDS = None

    # Select Image Bands if required
    inputImgBands = input_img
    selectBands = False
    if not bands is None:
        print("Subsetting the image bands")
        selectBands = True
        inputImgBands = os.path.join(
            tmp_dir, "{}_bselect.{}".format(basename, out_file_ext)
        )
        rsgislib.imageutils.select_img_bands(
            input_img, inputImgBands, gdalformat, input_datatype, bands
        )

    # Stretch input data if required.
    segmentFile = inputImgBands
    if not no_stretch:
        segmentFile = os.path.join(
            tmp_dir, "{}_stchd.{}".format(basename, out_file_ext)
        )
        strchFile = os.path.join(
            tmp_dir, "{}_stchdonly.{}".format(basename, out_file_ext)
        )
        strchFileOffset = os.path.join(
            tmp_dir, "{}_stchdonly_off.{}".format(basename, out_file_ext)
        )
        strchMaskFile = os.path.join(
            tmp_dir, "{}_stchdmaskonly.{}".format(basename, out_file_ext)
        )

        print("Stretch Input Image")
        rsgislib.imageutils.stretch_img_with_stats(
            inputImgBands,
            strchFile,
            img_stretch_stats,
            gdalformat,
            rsgislib.TYPE_8UINT,
            0,
            rsgislib.imageutils.STRETCH_LINEARMINMAX,
            2,
        )

        print(
            "Add 1 to stretched file to ensure there are no "
            "all zeros (i.e., no data) regions created."
        )
        rsgislib.imagecalc.image_math(
            strchFile, strchFileOffset, "b1+1", gdalformat, rsgislib.TYPE_8UINT
        )

        print("Create Input Image Mask.")
        bandMathBands = [
            rsgislib.imagecalc.BandDefn(
                band_name="b1", input_img=inputImgBands, img_band=1
            )
        ]
        rsgislib.imagecalc.band_math(
            strchMaskFile, "b1==0?1:0", gdalformat, rsgislib.TYPE_8UINT, bandMathBands
        )

        print("Mask stretched Image.")
        rsgislib.imageutils.mask_img(
            strchFileOffset,
            strchMaskFile,
            segmentFile,
            gdalformat,
            rsgislib.TYPE_8UINT,
            0,
            1,
        )

        if not no_delete:
            # Deleting extra files
            rsgislib.tools.filetools.delete_file_with_basename(strchFile)
            rsgislib.tools.filetools.delete_file_with_basename(strchFileOffset)
            rsgislib.tools.filetools.delete_file_with_basename(strchMaskFile)

    # Apply KMEANS
    print("Apply KMeans to image.")
    kMeansFileZones = os.path.join(
        tmp_dir, "{}_kmeans.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.label_pixels_from_cluster_centres(
        segmentFile, kMeansFileZones, kmeans_centres, True, gdalformat
    )

    # Elimininate Single Pixels
    print("Eliminate Single Pixels.")
    kMeansFileZonesNoSgls = os.path.join(
        tmp_dir, "{}_kmeans_nosgl.{}".format(basename, out_file_ext)
    )
    kMeansFileZonesNoSglsTmp = os.path.join(
        tmp_dir, "{}_kmeans_nosgl_tmp.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.eliminate_single_pixels(
        segmentFile,
        kMeansFileZones,
        kMeansFileZonesNoSgls,
        kMeansFileZonesNoSglsTmp,
        gdalformat,
        process_in_mem,
        True,
    )

    # Clump
    print("Perform clump.")
    initClumpsFile = os.path.join(
        tmp_dir, "{}_clumps.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.clump(
        kMeansFileZonesNoSgls, initClumpsFile, gdalformat, process_in_mem, 0
    )

    # Elimininate small clumps
    print("Eliminate small pixels.")
    elimClumpsFile = os.path.join(
        tmp_dir, "{}_clumps_elim.{}".format(basename, out_file_ext)
    )
    rsgislib.segmentation.rm_small_clumps_stepwise(
        segmentFile,
        initClumpsFile,
        elimClumpsFile,
        gdalformat,
        False,
        "",
        False,
        process_in_mem,
        min_n_pxls,
        dist_thres,
    )

    # Relabel clumps
    print("Relabel clumps.")
    rsgislib.segmentation.relabel_clumps(
        elimClumpsFile, out_clumps_img, gdalformat, process_in_mem
    )

    # Populate with stats if required.
    if calc_stats:
        print("Calculate image statistics and build pyramids.")
        rsgislib.rastergis.pop_rat_img_stats(out_clumps_img, True, True)

    # Create mean image if required.
    if not (out_mean_img is None):
        rsgislib.segmentation.mean_image(
            input_img, out_clumps_img, out_mean_img, gdalformat, input_datatype
        )
        if calc_stats:
            rsgislib.imageutils.pop_img_stats(out_mean_img, True, 0, True)

    if not no_delete:
        # Deleting extra files
        rsgislib.tools.filetools.delete_file_with_basename(kMeansFileZones)
        rsgislib.tools.filetools.delete_file_with_basename(kMeansFileZonesNoSgls)
        rsgislib.tools.filetools.delete_file_with_basename(kMeansFileZonesNoSglsTmp)
        rsgislib.tools.filetools.delete_file_with_basename(initClumpsFile)
        rsgislib.tools.filetools.delete_file_with_basename(elimClumpsFile)
        if selectBands:
            rsgislib.tools.filetools.delete_file_with_basename(inputImgBands)
        if not no_stretch:
            rsgislib.tools.filetools.delete_file_with_basename(segmentFile)
        if createdDIR:
            shutil.rmtree(tmp_dir)



#! /usr/bin/env python
############################################################################
#  classkeraschips.py
#
#  Copyright 2020 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of functions to apply tensorflow keras classifiers
#           to remotely sensed imagery.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 14/01/2020
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import rsgislib
import rsgislib.rastergis
import rsgislib.imagecalc

import tqdm

import numpy

import h5py

from osgeo import gdal

from rios.imagereader import ImageReader
from rios.imagewriter import ImageWriter
from rios import rat


def train_keras_chips_pixel_classifier(
    cls_mdl, cls_info_dict, out_mdl_file=None, train_epochs=5, train_batch_size=32
):
    """
    A function which trains a neural network defined using the keras API for the classification of remotely sensed data.
    A dict of class information, as ClassInfoObj objects, is defined with the training data.

    This function requires that tensorflow and keras modules to be installed.

    :param out_mdl_file: The output model which can be loaded to perform a classification.
    :param cls_info_dict: dict (key is string with class name) of ClassInfoObj objects defining the training data.
    :param out_mdl_file: A file path to save the trained model as a hdf5 file. If None then ignored.
    :param train_epochs: The number of epochs to use for training
    :param train_batch_size: The batch size to use for training.

    """
    try:
        from keras.utils import to_categorical
    except:
        from tensorflow.keras.utils import to_categorical

    n_classes = len(cls_info_dict)
    for clsname in cls_info_dict:
        if cls_info_dict[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, cls_info_dict[clsname].id
                )
            )

    cls_data_dict = {}
    train_data_lst = []
    train_lbls_lst = []
    valid_data_lst = []
    valid_lbls_lst = []
    test_data_lst = []
    test_lbls_lst = []
    cls_ids = []
    for clsname in cls_info_dict:
        sgl_cls_info = {}
        print("Reading Class {} Training".format(clsname))
        f = h5py.File(cls_info_dict[clsname].train_file_h5, "r")
        sgl_cls_info["train_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["train_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["train_data_lbls"] = numpy.zeros(
            sgl_cls_info["train_n_rows"], dtype=int
        )
        sgl_cls_info["train_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        train_data_lst.append(sgl_cls_info["train_data"])
        train_lbls_lst.append(sgl_cls_info["train_data_lbls"])

        print("Reading Class {} Validation".format(clsname))
        f = h5py.File(cls_info_dict[clsname].valid_file_h5, "r")
        sgl_cls_info["valid_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["valid_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["valid_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["valid_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        valid_data_lst.append(sgl_cls_info["valid_data"])
        valid_lbls_lst.append(sgl_cls_info["valid_data_lbls"])

        print("Reading Class {} Testing".format(clsname))
        f = h5py.File(cls_info_dict[clsname].test_file_h5, "r")
        sgl_cls_info["test_n_rows"] = f["DATA/DATA"].shape[0]
        sgl_cls_info["test_data"] = numpy.array(f["DATA/DATA"])
        sgl_cls_info["test_data_lbls"] = numpy.zeros(
            sgl_cls_info["valid_n_rows"], dtype=int
        )
        sgl_cls_info["test_data_lbls"][...] = cls_info_dict[clsname].id
        f.close()
        test_data_lst.append(sgl_cls_info["test_data"])
        test_lbls_lst.append(sgl_cls_info["test_data_lbls"])

        cls_data_dict[clsname] = sgl_cls_info
        cls_ids.append(cls_info_dict[clsname].id)

    print("Finished Reading Data")

    train_np = numpy.concatenate(train_data_lst)
    train_lbls_np = numpy.concatenate(train_lbls_lst)
    train_lbls_keras = to_categorical(train_lbls_np, num_classes=n_classes)

    vaild_np = numpy.concatenate(valid_data_lst)
    vaild_lbls_np = numpy.concatenate(valid_lbls_lst)
    vaild_lbls_keras = to_categorical(vaild_lbls_np, num_classes=n_classes)

    test_np = numpy.concatenate(test_data_lst)
    test_lbls_np = numpy.concatenate(test_lbls_lst)
    test_lbls_keras = to_categorical(test_lbls_np, num_classes=n_classes)

    print("Start Training Model")
    cls_mdl.fit(
        train_np,
        train_lbls_keras,
        epochs=train_epochs,
        batch_size=train_batch_size,
        validation_data=(vaild_np, vaild_lbls_keras),
    )
    print("Finished Training Model")
    cls_mdl.summary()

    loss_and_metrics = cls_mdl.evaluate(
        test_np, test_lbls_keras, batch_size=train_batch_size
    )
    eval_metric_names = cls_mdl.metrics_names
    for eval_name, eval_val in zip(eval_metric_names, loss_and_metrics):
        print("{} = {}".format(eval_name, eval_val))

    if out_mdl_file is not None:
        cls_mdl.save(out_mdl_file)


def apply_keras_chips_pixel_classifier(
    class_train_info,
    keras_cls_mdl,
    in_img_mask,
    img_mask_val,
    img_file_info,
    chip_size,
    out_class_img,
    norm_function=None,
    gdalformat="KEA",
    class_clr_names=True,
):
    """
    This function applies a trained single pixel keras model to an image. The function train_keras_pixel_classifier
    can be used to train such as model. The output image will contain the hard membership of the predicted class.


    :param class_train_info: dict (where the key is the class name) of rsgislib.classification.ClassInfoObj
                           objects which will be used to train the classifier (i.e., train_keras_pixel_classifier()),
                           provide pixel value id and RGB class values.
    :param keras_cls_mdl: a trained keras model object, with a input dimensions equivlent to the number of image
                          bands specified in the imgFileInfo input and output layer which provides an output array
                          of the length of the number of classes.
    :param in_img_mask: is an image file providing a mask to specify where should be classified. Simplest mask is all the
                    valid data regions (rsgislib.imageutils.gen_valid_mask)
    :param img_mask_val: the pixel value within the imgMask to limit the region to which the classification is applied.
                       Can be used to create a heirachical classification.
    :param img_file_info: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                        rsgislib.zonalstats.extract_zone_img_band_values_to_hdf) to identify which images and bands are to
                        be used for the classification so it adheres to the training data.
    :param out_class_img: Output image which will contain the hard classification.
    :param norm_function: Normalisation function to apply before running classification
    :param chip_size: is the chip size to be extracted.
    :param gdalformat: is the output image format - all GDAL supported formats are supported.
    :param class_clr_names: default is True and therefore a colour table will the colours specified in ClassInfoObj
                          and a class_names (from classTrainInfo) column will be added to the output file.

    """
    n_classes = len(class_train_info)
    cls_id_lut = numpy.zeros(n_classes)
    for clsname in class_train_info:
        if class_train_info[clsname].id >= n_classes:
            raise rsgislib.RSGISPyException(
                "ClassInfoObj '{}' id ({}) is not consecutive starting from 0.".format(
                    clsname, class_train_info[clsname].id
                )
            )
        cls_id_lut[class_train_info[clsname].id] = class_train_info[clsname].out_id

    inImgs = list()
    inImgBands = list()

    inImgs.append(in_img_mask)
    inImgBands.append([1])
    n_img_bands = 0
    for inImgInfo in img_file_info:
        inImgs.append(inImgInfo.fileName)
        inImgBands.append(inImgInfo.bands)
        n_img_bands = n_img_bands + len(inImgInfo.bands)
    nImgs = len(img_file_info)

    scn_overlap = chip_size // 2

    writer = None
    reader = ImageReader(
        inImgs,
        windowxsize=chip_size,
        windowysize=chip_size,
        overlap=scn_overlap,
        layerselection=inImgBands,
    )
    feat2cls = None

    for (info, block) in tqdm.tqdm(reader):
        classMskArr = block[0]

        blk_shape = classMskArr.shape

        if feat2cls is None:
            feat2cls = numpy.zeros(
                [1, chip_size, chip_size, n_img_bands], dtype=numpy.float32
            )
        else:
            feat2cls[...] = 0

        chip_mask = classMskArr[
            0,
            scn_overlap : scn_overlap + chip_size,
            scn_overlap : scn_overlap + chip_size,
        ]
        # If there are no valid pixels in the block at all then no point running prediction
        # set all output valies to 0.
        if numpy.all(chip_mask != img_mask_val):
            out_cls_arr = numpy.zeros_like(classMskArr, dtype=numpy.uint16)
            out_cls_arr[...] = 0
        else:
            # Go through bands and reorder for prediction
            for nImg in range(nImgs):
                imgBlk = block[nImg + 1][:, :, :]
                for iBand in range(imgBlk.shape[0]):
                    feat2cls[0, :, :, iBand] = imgBlk[
                        iBand,
                        scn_overlap : scn_overlap + chip_size,
                        scn_overlap : scn_overlap + chip_size,
                    ]

            image_chip = feat2cls[0:1, :, :, :]
            # If a normalisation function has been specified then apply this
            if norm_function is not None:
                image_chip = norm_function(image_chip)

            predict_class = numpy.argmax(keras_cls_mdl.predict(image_chip), axis=1)

            # Update based on output class
            out_predict_class = numpy.empty_like(predict_class)
            for cld_id, idx in zip(cls_id_lut, numpy.arange(0, len(cls_id_lut))):
                out_predict_class[predict_class == idx] = cld_id

            # Assign all pixels in block to predicted class
            out_cls_arr = numpy.zeros_like(classMskArr, dtype=numpy.uint16)
            out_cls_arr[...] = out_predict_class

            # Apply mask to output pixels
            out_cls_arr[classMskArr != img_mask_val] = 0

        if writer is None:
            writer = ImageWriter(
                out_class_img, info=info, firstblock=out_cls_arr, drivername=gdalformat
            )
        else:
            writer.write(out_cls_arr)
    writer.close(calcStats=False)

    if class_clr_names:
        rsgislib.rastergis.pop_rat_img_stats(
            out_class_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )
        max_val = rsgislib.imagecalc.get_img_band_min_max(out_class_img, 1, False, 0)[1]
        ratDataset = gdal.Open(out_class_img, gdal.GA_Update)

        max_cls_val = 0
        for classKey in class_train_info:
            if class_train_info[classKey].out_id > max_cls_val:
                max_cls_val = class_train_info[classKey].out_id

        if max_cls_val > max_val:
            red = numpy.random.randint(0, 255, max_cls_val + 1)
            green = numpy.random.randint(0, 255, max_cls_val + 1)
            blue = numpy.random.randint(0, 255, max_cls_val + 1)
        else:
            red = rat.readColumn(ratDataset, "Red")
            green = rat.readColumn(ratDataset, "Green")
            blue = rat.readColumn(ratDataset, "Blue")

        class_names = numpy.empty_like(red, dtype=numpy.dtype("a255"))
        class_names[...] = ""

        for classKey in class_train_info:
            print("Apply Colour to class '" + classKey + "'")
            red[class_train_info[classKey].out_id] = class_train_info[classKey].red
            green[class_train_info[classKey].out_id] = class_train_info[classKey].green
            blue[class_train_info[classKey].out_id] = class_train_info[classKey].blue
            class_names[class_train_info[classKey].out_id] = classKey

        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)
        rat.writeColumn(ratDataset, "class_names", class_names)
        ratDataset = None


def train_keras_chips_ref_classifier(
    cls_mdl,
    train_data_file,
    valid_data_file,
    test_data_file,
    n_classes,
    out_mdl_file=None,
    train_epochs=5,
    train_batch_size=32,
):
    """
    A function which trains a neural network defined using the keras API for the classification of remotely sensed data.

    This function requires that tensorflow and keras modules to be installed.

    :param cls_mdl: The Keras model which has already been defined.
    :param train_data_file: a HDF5 file with training data (needs to have elements /DATA/DATA and /DATA/REF)
    :param valid_data_file: a HDF5 file with validation data (needs to have elements /DATA/DATA and /DATA/REF)
    :param test_data_file: a HDF5 file with testing data (needs to have elements /DATA/DATA and /DATA/REF)
    :param n_classes: integer specifying the number of classes within the system
    :param out_mdl_file: A file path to save the trained model as a hdf5 file. If None then ignored.
    :param train_epochs: The number of epochs to use for training
    :param train_batch_size: The batch size to use for training.

    """
    try:
        from keras.utils import to_categorical
    except:
        from tensorflow.keras.utils import to_categorical

    f = h5py.File(train_data_file, "r")
    train_np = numpy.array(f["DATA/DATA"])
    train_lbls_np = numpy.array(f["DATA/REF"])
    f.close()
    train_lbls_keras = to_categorical(train_lbls_np, num_classes=n_classes)

    f = h5py.File(valid_data_file, "r")
    vaild_np = numpy.array(f["DATA/DATA"])
    vaild_lbls_np = numpy.array(f["DATA/REF"])
    f.close()
    vaild_lbls_keras = to_categorical(vaild_lbls_np, num_classes=n_classes)

    f = h5py.File(test_data_file, "r")
    test_np = numpy.array(f["DATA/DATA"])
    test_lbls_np = numpy.array(f["DATA/REF"])
    f.close()
    test_lbls_keras = to_categorical(test_lbls_np, num_classes=n_classes)
    print("Finished Reading Data")

    print("Start Training Model")
    cls_mdl.fit(
        train_np,
        train_lbls_keras,
        epochs=train_epochs,
        batch_size=train_batch_size,
        validation_data=(vaild_np, vaild_lbls_keras),
    )
    print("Finished Training Model")
    cls_mdl.summary()

    if out_mdl_file is not None:
        cls_mdl.save(out_mdl_file)

    loss_and_metrics = cls_mdl.evaluate(
        test_np, test_lbls_keras, batch_size=train_batch_size
    )
    eval_metric_names = cls_mdl.metrics_names
    for eval_name, eval_val in zip(eval_metric_names, loss_and_metrics):
        print("{} = {}".format(eval_name, eval_val))



#!/usr/bin/env python
"""
The image registration module contains algorithms for generating tie points matching two image and warping images based on tie points.

There are two algorithms are available for registration: basic, and singlelayer. The single layer algorithm is a simplified version of the algorithm proposed in:

Bunting, P.J., Labrosse, F. & Lucas, R.M., 2010. A multi-resolution area-based technique for automatic multi-modal image registration. Image and Vision Computing, 28(8), pp.1203-1219.


Image distance metrics:

    * METRIC_EUCLIDEAN = 1
    * METRIC_SQDIFF = 2
    * METRIC_MANHATTEN = 3
    * METRIC_CORELATION = 4


GCP Output Types:

    * TYPE_ENVI_IMG2IMG = 1
    * TYPE_ENVI_IMG2MAP = 2
    * TYPE_RSGIS_IMG2MAP = 3
    * TYPE_RSGIS_MAPOFFS = 4


"""

# import the C++ extension into this level
from ._imageregistration import *

import rsgislib

METRIC_EUCLIDEAN = 1
METRIC_SQDIFF = 2
METRIC_MANHATTEN = 3
METRIC_CORELATION = 4

TYPE_ENVI_IMG2IMG = 1
TYPE_ENVI_IMG2MAP = 2
TYPE_RSGIS_IMG2MAP = 3
TYPE_RSGIS_MAPOFFS = 4


def warp_with_gcps_with_gdal(
    in_ref_img: str,
    in_process_img: str,
    output_img: str,
    gdalformat: str,
    interp_method: int = rsgislib.INTERP_NEAREST_NEIGHBOUR,
    use_tps: bool = False,
    use_poly: bool = True,
    poly_order: int = 3,
    use_multi_thread: bool = False,
):
    """
    A utility function to warp an image file (in_process_img) using GCPs defined within
    the image header - this is the same as using the gdalwarp utility. However, the
    output image will have the same pixel grid and dimensions as the input reference
    image (in_ref_img).

    :param in_ref_img: is the input reference image to which the processing image
                       is to resampled to.
    :param in_process_img: is the image which is to be resampled.
    :param output_img: is the output image file.
    :param gdalformat: is the gdal format for the output image.
    :param interp_method: is the interpolation method used to resample the image
                          [bilinear, lanczos, cubicspline, nearestneighbour, cubic,
                          average, mode]
    :param use_tps: is a boolean specifying that the thin plated splines method of
                    warping should be used (i.e., rubbersheet); Default False.
    :param use_poly: is a boolean specifying that a polynomial method of warping
                     is used; Default True
    :param poly_order: is the order of the polynomial used to represent the
                       transformation (1, 2 or 3). Only used if use_poly=True
    :param use_multi_thread: is a boolean specifying whether multiple processing
                             cores should be used for the warping.

    """
    import rsgislib.imageutils
    from osgeo import gdal

    if not rsgislib.imageutils.has_gcps(in_process_img):
        raise rsgislib.RSGISPyException(
            "Input process image does not have GCPs "
            "within it's header - this is required."
        )

    numBands = rsgislib.imageutils.get_img_band_count(in_process_img)
    noDataVal = rsgislib.imageutils.get_img_no_data_value(in_process_img)
    datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(in_process_img)

    interpolationMethod = gdal.GRA_NearestNeighbour
    if interp_method == rsgislib.INTERP_BILINEAR:
        interpolationMethod = gdal.GRA_Bilinear
    elif interp_method == rsgislib.INTERP_LANCZOS:
        interpolationMethod = gdal.GRA_Lanczos
    elif interp_method == rsgislib.INTERP_CUBICSPLINE:
        interpolationMethod = gdal.GRA_CubicSpline
    elif interp_method == rsgislib.INTERP_NEAREST_NEIGHBOUR:
        interpolationMethod = gdal.GRA_NearestNeighbour
    elif interp_method == rsgislib.INTERP_CUBIC:
        interpolationMethod = gdal.GRA_Cubic
    elif interp_method == rsgislib.INTERP_AVERAGE:
        interpolationMethod = gdal.GRA_Average
    elif interp_method == rsgislib.INTERP_MODE:
        interpolationMethod = gdal.GRA_Mode
    else:
        raise rsgislib.RSGISPyException(
            "Interpolation method was not recognised or known."
        )

    rsgislib.imageutils.create_copy_img(
        in_ref_img, output_img, numBands, 0, gdalformat, datatype
    )

    inFile = gdal.Open(in_process_img, gdal.GA_ReadOnly)
    outFile = gdal.Open(output_img, gdal.GA_Update)

    try:
        import tqdm

        pbar = tqdm.tqdm(total=100)
        callback = lambda *args, **kw: pbar.update()
    except:
        callback = gdal.TermProgress

    wrpOpts = None
    if use_tps:
        wrpOpts = gdal.WarpOptions(
            resampleAlg=interpolationMethod,
            srcNodata=noDataVal,
            dstNodata=noDataVal,
            multithread=use_multi_thread,
            tps=use_tps,
            callback=callback,
        )
    elif use_poly:
        wrpOpts = gdal.WarpOptions(
            resampleAlg=interpolationMethod,
            srcNodata=noDataVal,
            dstNodata=noDataVal,
            multithread=use_multi_thread,
            polynomialOrder=poly_order,
            callback=callback,
        )
    else:
        warpOptions = None

    gdal.Warp(outFile, inFile, options=wrpOpts)

    inFile = None
    outFile = None
